@@ -1,13 +1,17 @@
 #!/usr/bin/env perl
 
+use 5.026;    # indented heredocs
 use strict;
 use warnings;
 use experimental 'signatures';
+use feature 'state';
 use YAML::XS qw(LoadFile);
 use Path::Tiny;
 use JSON::PP;
 use Carp  qw(croak);
 use Clone qw(clone);
+use Template;
+use Markdown::Pod;
 
 # FIXME: switch from HEREDOCs to Templates
 
@@ -46,22 +50,24 @@ sub preprocess_openapi ($openapi) {
 
 # walks through the OpenAPI spec and resolves all references.
 sub _recursively_find_references ( $components, $resolved ) {
+
+    # Cache the references to avoid resolving them multiple times. The script
+    # was fast enough and memory efficient enough to not really need this, but
+    # we cut memory usage by more than half and improved speed an order of
+    # magnitude.
+    state $cached = {};
     return unless ref $resolved;
     if ( 'ARRAY' eq ref $resolved ) {
         foreach my $item ( $resolved->@* ) {
             _recursively_find_references( $components, $item );
         }
     } elsif ( 'HASH' eq ref $resolved ) {
-        if ( exists $resolved->{'$ref'} ) {
-            my $reference = _resolve_reference( $components, delete $resolved->{'$ref'} );
+        if ( my $ref = $resolved->{'$ref'} ) {
+            my $reference = $cached->{$ref} //= _resolve_reference( $components, $ref );
             $resolved->%* = ( %{$reference}, %{$resolved} );    # Merge reference into current hash
         }
     KEY: foreach my $key ( sort keys $resolved->%* ) {
             my $item = $resolved->{$key};
-            if ( 'x-oaiMeta' eq $key ) {
-                delete $resolved->{$key};
-                next KEY;
-            }
             _recursively_find_references( $components, $item );
         }
     }
@@ -73,15 +79,29 @@ sub _resolve_reference ( $components, $ref ) {
 }
 
 sub get_example_from_schema ($schema) {
+    if ( my $example = $schema->{'x-oaiMeta'}{example} ) {
+
+        # FIXME: this is a nasty, nasty hack. x-oaiMeta is not part of the
+        # OpenAPI spec, but it gives much better examples than trying to parse
+        # manually. However, while those examples are easier to read and more
+        # comprehensive, they're often not valid JSON. We return them as-is
+        # and let the calling code handle it. Because this is a *recursive*
+        # function, it's fragile, but so far, we only fine the x-oaiMeta key
+        # at the top level, so we never hit recursion.
+
+        return $example;
+    }
     if ( defined $schema->{example} ) {
         return $schema->{example};
     } elsif ( defined $schema->{properties} ) {
         my %example;
-        for my $property_name ( keys %{ $schema->{properties} } ) {
+    PROPERTY: foreach my $property_name ( keys %{ $schema->{properties} } ) {
             my $property_schema = $schema->{properties}->{$property_name};
             if ( defined $property_schema->{example} ) {
                 $example{$property_name} = $property_schema->{example};
-            } elsif ( defined $property_schema->{type}
+                next PROPERTY;
+            }
+            if (   defined $property_schema->{type}
                 && $property_schema->{type} eq 'array'
                 && defined $property_schema->{items} ) {
                 $example{$property_name} = [ get_example_from_schema( $property_schema->{items} ) ];
@@ -98,6 +118,16 @@ sub get_example_from_schema ($schema) {
     return undef;
 }
 
+sub markdown_to_pod ($markdown) {
+    my $m2p = Markdown::Pod->new;
+
+    # FIXME: quick-n-dirty hack to convert links to the OpenAI docs
+    if ( $markdown =~ m{\(/docs/} ) {
+        $markdown =~ s{\(/docs/}{(https://platform.openai.com/docs/}g;
+    }
+    return $m2p->markdown_to_pod( markdown => $markdown );
+}
+
 sub write_documentation_for_path ( $resolved_openapi, $path, $output_base_dir, $path_index_entries ) {
     my $path_data              = $resolved_openapi->{paths}->{$path};
     my $sanitized_path_segment = $path;
@@ -106,111 +136,61 @@ sub write_documentation_for_path ( $resolved_openapi, $path, $output_base_dir, $
     $sanitized_path_segment =~ s{-+$}{};
     $sanitized_path_segment =~ s{--}{-}g;
 
-    my $output_file = $output_base_dir->child("$sanitized_path_segment.pod");
-
     eval {
-        $output_file->spew_utf8(<<~"POD");
-          =encoding utf8
-          
-          =head1 NAME
-          
-          OpenAPI::Client::OpenAI::Path::$sanitized_path_segment - Documentation for the $path path.
-          
-          =head1 DESCRIPTION
-          
-          This document describes the API endpoint at C<$path>.
+        my $tt       = Template->new( { TRIM => 1 } );
+        my $template = _path_template();
+        my $json     = JSON::PP->new->pretty->canonical;
 
-          See the C<examples/> directory in the distribution for examples of how to use this.
-          
-          POD
-
-        if ( defined $path_data->{description} && $path_data->{description} ne '' ) {
-            $output_file->append("$path_data->{description}\n\n");
-        }
-
-        my $json = JSON::PP->new->pretty;
+        my %template_data = (
+            path                   => $path,
+            sanitized_path_segment => $sanitized_path_segment,
+            path_data              => $path_data,
+            year                   => (localtime)[5] + 1900,
+        );
         foreach my $method ( sort keys %{$path_data} ) {
             next if $method eq 'description' || $method eq 'parameters';
 
-            my $method_data  = $path_data->{$method};
-            my $method_upper = uc $method;
-            my $operation_id = $method_data->{operationId};
-
-            $output_file->append("=head2 C<$method_upper $path>\n\n");
+            my $method_data = $path_data->{$method};
 
             if ( defined $method_data->{summary} && $method_data->{summary} ne '' ) {
-                $output_file->append("$method_data->{summary}\n\n");
+                $method_data->{summary} = markdown_to_pod( $method_data->{summary} );
             }
             if ( defined $method_data->{description} && $method_data->{description} ne '' ) {
-                $output_file->append("$method_data->{description}\n");
-            }
-
-            if (defined $operation_id) {
-                $output_file->append(<<~"POD");
-                  =head3 Operation ID
-                  
-                  C<$operation_id>
-
-                      \$client->$operation_id( ... );
-                  
-                  POD
-            }
-
-            # Add parameter documentation
-            if ( defined $method_data->{parameters} && @{ $method_data->{parameters} } ) {
-                $output_file->append(<<~'POD');
-                  =head3 Parameters
-                  
-                  =over 4
-                  
-                  POD
-
-                foreach my $parameter ( @{ $method_data->{parameters} } ) {
-                    my $name        = $parameter->{name};
-                    my $in          = $parameter->{in};
-                    my $description = $parameter->{description} || 'No description available.';
-                    my $required    = $parameter->{required} ? '(Required)' : '(Optional)';
-
-                    $output_file->append("=item * C<$name> (in $in) $required - $description\n");
-
-                    if ( defined $parameter->{schema} && defined $parameter->{schema}->{type} ) {
-                        $output_file->append("Type: C<$parameter->{schema}->{type}>\n\n");
-
-                        if ( defined $parameter->{schema}->{format} ) {
-                            $output_file->append("Format: C<$parameter->{schema}->{format}>\n");
-
-                        }
-                        if ( defined $parameter->{schema}->{enum} && @{ $parameter->{schema}->{enum} } ) {
-                            $output_file->append("Possible values: C<@{ $parameter->{schema}->{enum} }>\n\n");
-                        }
-                        if ( defined $parameter->{schema}->{default} ) {
-                            $output_file->append("Default: C<$parameter->{schema}->{default}>\n\n");
-                        }
-                    }
-                }
-                $output_file->append("\n=back\n\n");
+                $method_data->{description} = markdown_to_pod( $method_data->{description} );
             }
 
             # Add request body documentation with examples
             if ( defined $method_data->{requestBody} && defined $method_data->{requestBody}->{content} ) {
-                $output_file->append("\n=head3 Request Body\n\n");
-                foreach my $content_type ( sort keys %{ $method_data->{requestBody}->{content} } ) {
-                    $output_file->append("=head3 Content Type: C<$content_type>\n\n");
-
-                    if ( defined $method_data->{requestBody}->{content}->{$content_type}->{schema} ) {
-                        my $schema = $method_data->{requestBody}->{content}->{$content_type}->{schema};
+                CONTENT_TYPE: foreach my $content_type ( sort keys %{ $method_data->{requestBody}->{content} } ) {
 
+                    if ( defined $method_data->{requestBody}{content}{$content_type}{schema} ) {
+                        my $schema = $method_data->{requestBody}{content}{$content_type}{schema};
                         if ( defined( my $example = get_example_from_schema($schema) ) ) {
                             my $example_json = $json->encode($example);
                             # prepend each line with four spaces
                             $example_json =~ s/^/    /gm;
-                            $output_file->append(<<~"POD");
-                              
-                              =head3 Example:
-                              
-                              $example_json
-                              
-                              POD
+                            $schema->{example} = $example_json;
+                        }
+
+                        # Try to grab the model data
+                        if ( my $properties = $schema->{properties} ) {
+                            my $model = $properties->{model} or next CONTENT_TYPE;
+                            if ( my $description = $model->{description} ) {
+                                my $description = markdown_to_pod( $model->{description} );
+                                my $models;
+                            ELEM: foreach my $elem ( @{ $model->{anyOf} } ) {
+                                    if ( defined $elem->{type} && exists $elem->{enum} ) {
+                                        $models = $elem->{enum};
+                                        last;
+                                    }
+                                }
+                                if ($models) {
+                                    $schema->{models} = {
+                                        description => $description,
+                                        models      => $models,
+                                    };
+                                }
+                            }
                         }
                     }
                 }
@@ -218,59 +198,38 @@ sub write_documentation_for_path ( $resolved_openapi, $path, $output_base_dir, $
 
             # Add responses documentation with examples
             if ( defined $method_data->{responses} ) {
-                $output_file->append("\n=head3 Responses\n\n");
-
                 foreach my $status_code ( sort keys %{ $method_data->{responses} } ) {
                     my $response = $method_data->{responses}->{$status_code};
-                    $output_file->append("=head3 Status Code: C<$status_code>\n\n");
 
-                    if ( defined $response->{description} ) {
-                        $output_file->append("$response->{description}\n\n");
-                    }
                     if ( defined $response->{content} ) {
-                        $output_file->append("Content Types:\n\n=over 4\n\n");
                         foreach my $content_type ( sort keys %{ $response->{content} } ) {
-                            $output_file->append("=item * C<$content_type>\n");
-                            if ( defined $response->{content}->{$content_type}->{schema} ) {
-                                my $schema = $response->{content}->{$content_type}->{schema};
+                            if ( defined $response->{content}{$content_type}{schema} ) {
+                                my $schema = $response->{content}{$content_type}{schema};
 
                                 if ( defined( my $example = get_example_from_schema($schema) ) ) {
-                                    my $example_json = $json->encode($example);
+                                    my $example_json = ref $example ? $json->encode($example) : $example;
                                     # prepend each line with four spaces
                                     $example_json =~ s/^/    /gm;
-                                    $output_file->append(<<~"POD");
-                                      
-                                      Example:
-                                      
-                                      $example_json
-                                      
-                                      POD
+                                    $schema->{example} = $example_json;
                                 }
                             }
                         }
-                        $output_file->append("\n=back\n\n");
                     }
                 }
             }
         }
-        $output_file->append(<<POD);
-
-=head1 SEE ALSO
-
-L<OpenAPI::Client::OpenAI::Path>
-
-=head1 COPYRIGHT AND LICENSE
-
-Copyright (C) 2023-2025 by Nelson Ferraz
-
-This library is free software; you can redistribute it and/or modify
-it under the same terms as Perl itself, either Perl version 5.14.0 or,
-at your option, any later version of Perl 5 you may have available.
-
-=cut
-POD
+        my $output;
+        $tt->process( \$template, \%template_data, \$output )
+            or die "Template processing failed: " . $tt->error;
+        #warn $template_data{path_data}{get}{responses}{200}{content}{'application/json'}{schema}{example};
+        my $output_file = $output_base_dir->child("$sanitized_path_segment.pod");
+        $output_file->spew_utf8($output);
+        1;
+    } or do {
+        my $error = $@ || 'Unknown error';
+        die "Failed to write documentation for path '$path': $error\n";
     };
-    die "Error writing to '$output_file': $@" if $@;
+
 
     # Store information for the main index
     $path_index_entries->{$path} = {
@@ -336,3 +295,121 @@ sub write_index ( $main_index_file, $path_index_entries, $resolved_openapi ) {
 
     close $index_fh;
 }
+
+sub _path_template () {
+    return <<~'END_TEMPLATE';
+    =encoding utf8
+    
+    =head1 NAME
+    
+    OpenAPI::Client::OpenAI::Path::[% sanitized_path_segment %] - Documentation for the [% path %] path.
+    
+    =head1 DESCRIPTION
+    
+    This document describes the API endpoint at C<[% path %]>.
+
+    =head1 PATHS
+    
+    [% FOREACH http_method IN path_data.keys.sort %]
+    =head2 C<[% http_method.upper %] [% path %]>
+
+    [% method_data = path_data.$http_method; method_data.summary %]
+    [% IF method_data.description %]
+    [% method_data.description %]
+    [% END %]
+    
+    =head3 Operation ID
+    
+    C<[% method_data.operationId %]>
+    
+        $client->[% method_data.operationId %]( ... );
+    
+    =head3 Parameters
+    
+    =over 4
+    [% FOREACH parameter IN method_data.parameters %]
+    =item * C<[% parameter.name %]> (in [% parameter.in %]) [% IF parameter.required %](Required)[% ELSE %](Optional)[% END %] - [% parameter.description %]
+
+    Type: C<[% parameter.schema.type %]>
+    [% IF parameter.schema.enum.size %]
+    Allowed values: C<[% parameter.schema.enum.join(', ') %]>
+    [% END %]
+    [% IF parameter.schema.default %]
+    Default: C<[% parameter.schema.default %]>
+    [% END -%]
+    
+    [% END # END FOREACH %]
+    =back
+    [% IF method_data.requestBody %]
+    =head3 Request Body
+      [% FOREACH content_type IN method_data.requestBody.content.keys.sort %]
+    =head3 Content Type: C<[% content_type %]>
+    [% schema = method_data.requestBody.content.$content_type.schema -%]
+    
+        [% IF schema %]
+          [% schema.description %]
+    
+          [% IF schema.models %]
+
+    =head4 Models
+
+    [% schema.models.description %]
+    =over 4
+          [% FOREACH model IN schema.models.models %]
+    =item * C<[% model %]>
+    [% END # FOREACH model %]
+    =back
+          [% END # IF models -%]
+    
+          [% IF schema.example %]
+    Example:
+    
+    [% schema.example %]
+    
+          [% END # IF example -%]
+        [% END # end if requestBody.content -%]
+       [% END # FOREACH content_type -%]
+    [% END # end if method_data.requestBody -%]
+    
+    [% IF method_data.responses %]
+    =head3 Responses
+    
+    [% FOREACH status_code IN method_data.responses.keys.sort %]
+    =head4 Status Code: C<[% status_code %]>
+    
+    [% method_data.responses.$status_code.description %]
+    
+    [% IF method_data.responses.$status_code.content %]
+    =head4 Content Types:
+    
+    =over 4
+    
+    [% FOREACH content_type IN method_data.responses.$status_code.content.keys.sort %]
+    =item * C<[% content_type %]>
+    
+    Example (See the L<OpenAI spec for more detail|https://github.com/openai/openai-openapi/blob/master/openapi.yaml>):
+    
+    [% method_data.responses.$status_code.content.$content_type.schema.example %]
+    
+    [% END # FOREACH content_type -%]
+    =back
+    [% END # end if status_code.content -%]
+    [% END # end FOREACH status_code -%]
+    [% END # if method_data.resposnes -%] 
+    [% END # end FOREACH http_method -%]
+    
+    =head1 SEE ALSO
+    
+    L<OpenAPI::Client::OpenAI::Path>
+    
+    =head1 COPYRIGHT AND LICENSE
+    
+    Copyright (C) 2023-[% year %] by Nelson Ferraz
+    
+    This library is free software; you can redistribute it and/or modify
+    it under the same terms as Perl itself, either Perl version 5.14.0 or,
+    at your option, any later version of Perl 5 you may have available.
+    
+    =cut
+    END_TEMPLATE
+}
